#coding: utf-8
from collections import OrderedDict


class Dicionario(OrderedDict):
    """
    Dicionário especializado que permite acesso de chaves como propriedades.
    Adicionalmente, se uma chave já foi atribuída, transforma o valor em uma
    lista e acumula os valores.
    """
    def __getattr__(self, nome):
        if nome in self:
            return self[nome]
        raise AttributeError("%r sem atributo %r" %
                             (type(self).__name__, nome))



#coding: utf-8


class Cache(object):
    def __init__(self, tamanho=0):
        self.tamanho = tamanho
        self.lista = []

    def __len__(self):
        return len(self.lista)

    def adicionar(self, item):
        self.lista.append(item)
        if self.tamanho:
            while len(self.lista) > self.tamanho:
                self.lista.pop(0)

    def itens(self):
        return self.lista

    def consumir(self):
        while self.lista:
            yield self.lista.pop(0)



#coding: utf-8

"""
Os campos são simples extratores de dados baseados em expressões regulares.

Ao confrontar uma linha recebida para análise com sua expressão regular, o
campo verifica se há grupos de dados capturados, e então pode realizar algum
processamento e validações nestes dados. Se o campo considerar os dados
válidos, retorna o(s) dado(s). """

import re
from datetime import datetime
import collections


class CampoBase(object):
    """
    Contém lógica de processamento para extrair dados através de expressões
    regulares, além de prover métodos utilitários que podem ser sobrescritos
    para customizações no tratamento dos dados.

    O comportamento do Campo pode ser ajustado através de diversos parâmetros:

    mascara

        O requisito mínimo para um campo é uma máscara em expressão regular,
        onde deve-se especificar um grupo para captura::

            >>> s = "02/01/2013 10:21:51           COO:022734"
            >>> campo = CampoBase(mascara=r'COO:(\d+)')
            >>> campo.analizar_linha(s)
            '022734'

        O parâmetro mascara é o único posicional, e deste modo, seu nome pode
        ser omitido::

            >>> s = "02/01/2013 10:21:51           COO:022734"
            >>> campo = CampoBase(r'COO:(\d+)')
            >>> campo.analizar_linha(s)
            '022734'


    ao_atribuir

        Recebe um callback para tratar o valor antes de ser retornado pelo
        campo.

            >>> s = "02/01/2013 10:21:51           COO:022734"
            >>> def dobro(valor):
            ...     return int(valor) * 2
            ...
            >>> campo = CampoBase(r'COO:(\d+)', ao_atribuir=dobro)
            >>> campo.analizar_linha(s)  # 45468 = 2 x 22734
            45468

    grupos

        Permite escolher quais grupos capturados o campo deve processar como
        dados de entrada, utilizado para expressões regulares que utilizam
        grupos para correspondência da expressão regular, mas que apenas parte
        destes grupos possui informação útil.

        Pode-se informar um número inteiro, que será o índice do grupo,
        inicando em 0::

            >>> s = "Contador de Reduções Z:                     1246"
            >>> campo = CampoBase(r'Contador de Reduç(ão|ões) Z:\s*(\d+)', \
                grupos=1, ao_atribuir=int)
            >>> campo.analizar_linha(s)
            1246

        Ou uma lista de inteiros::

            >>> s = "Data do movimento: 02/01/2013 10:21:51"
            >>> c = CampoBase(r'^Data .*(movimento|cupom): (\d+)/(\d+)/(\d+)',\
                grupos=[1, 2, 3])
            >>> c.analizar_linha(s)
            ['02', '01', '2013']


    valor_padrao

        Valor que será utilizado no :py:class:`~raspador.analizador.Analizador`
        , quando o campo não retornar valor após a análise das
        linhas recebidas.


    lista

        Quando especificado, retorna o valor como uma lista::

            >>> s = "02/01/2013 10:21:51           COO:022734"
            >>> campo = CampoBase(r'COO:(\d+)', lista=True)
            >>> campo.analizar_linha(s)
            ['022734']

        Por convenção, quando um campo retorna uma lista, o
        :py:class:`~raspador.analizador.Analizador` acumula os valores
        retornados pelo campo.
    """
    def __init__(self, mascara=None, **kwargs):
        class_unique_name = self.__class__.__name__ + str(id(self))
        if not hasattr(self, 'nome'):
            self.nome = kwargs.get('nome', class_unique_name)

        self.valor_padrao = kwargs.get('valor_padrao')
        self.lista = kwargs.get('lista', False)
        self.mascara = re.compile(mascara) if mascara else None
        self.ao_atribuir = kwargs.get('ao_atribuir')
        self.grupos = kwargs.get('grupos', [])

        if self.ao_atribuir and \
                not isinstance(self.ao_atribuir, collections.Callable):
            raise TypeError('O callback ao_atribuir não é uma função.')

        if not hasattr(self.grupos, '__iter__'):
            self.grupos = (self.grupos,)

        self._iniciar()

    @property
    def _metodo_busca(self):
        return self.mascara.findall

    def _iniciar(self):
        "Ponto para inicialização especial nas classes descendentes"
        pass

    def atribuir_analizador(self, analizador):
        """
        Recebe uma referência fraca de
        :py:class:`~raspador.analizador.Analizador`
        """
        self.analizador = analizador

    def _resultado_valido(self, valor):
        return bool(valor)

    def _converter(self, valor):
        if self.grupos:
            if len(valor) == 1:   # não é desejado uma tupla,
                valor = valor[0]  # se houver apenas um item
            tamanho = len(valor)
            valor = [valor[i] for i in self.grupos if i < tamanho]

        if len(valor) == 1:   # não é desejado uma tupla,
            valor = valor[0]  # se houver apenas um item
        return valor

    def _para_python(self, valor):
        """
        Converte o valor recebido palo parser para o tipo de dado
        nativo do python
        """
        return valor

    @property
    def mascara(self):
        return self._mascara

    @mascara.setter
    def mascara(self, valor):
        self._mascara = re.compile(valor) if valor else None

    def anexar_na_classe(self, cls, nome, informacoes):
        self.cls = cls
        self.nome = nome
        self.informacoes = informacoes

    def analizar_linha(self, linha):
        if self.mascara:
            valor = self._metodo_busca(linha)
            if self._resultado_valido(valor):
                valor = self._converter(valor)
                valor = self._para_python(valor)
                if self.ao_atribuir:
                    valor = self.ao_atribuir(valor)
                if valor is not None and self.lista \
                        and not isinstance(valor, list):
                    valor = [valor]
                return valor


class CampoString(CampoBase):
    def _para_python(self, valor):
        return str(valor).strip()


class CampoNumerico(CampoBase):
    def _para_python(self, valor):
        valor = valor.replace('.', '')
        valor = valor.replace(',', '.')
        return float(valor)


class CampoInteiro(CampoBase):
    def _para_python(self, valor):
        return int(valor)


class CampoBooleano(CampoBase):
    """
    Retorna verdadeiro se a Regex bater com uma linha completa, e
    se ao menos algum valor for capturado.
    """
    def _iniciar(self):
        self.valor_padrao = False

    @property
    def _metodo_busca(self):
        return self.mascara.match

    def _converter(self, valor):
        res = valor.groups() if valor else False
        return super(CampoBooleano, self)._converter(res)

    def _resultado_valido(self, valor):
        return valor and (valor.groups())

    def _para_python(self, valor):
        return bool(valor)


class CampoData(CampoBase):
    """
    Campo que mantém dados no formato de data,
    representado em Python por datetine.date.

    Formato:
        Veja http://docs.python.org/library/datetime.html para detalhes.
    """

    def __init__(self, mascara=None, **kwargs):
        """
        formato='%d/%m/%Y'
        """
        super(CampoData, self).__init__(mascara=mascara, **kwargs)
        self.formato = kwargs.get('formato', '%d/%m/%Y')

    def _para_python(self, valor):
        return datetime.strptime(valor, self.formato).date()


class CampoDataHora(CampoBase):
    """
    Campo que mantém dados no formato de data/hora,
    representado em Python por datetine.datetime.

    Formato:
        Veja http://docs.python.org/library/datetime.html para detalhes.
    """

    def __init__(self, mascara=None, **kwargs):
        """
        formato='%d/%m/%Y %H:%M:%S'
        """
        super(CampoDataHora, self).__init__(mascara=mascara, **kwargs)
        self.formato = kwargs.get('formato', '%d/%m/%Y %H:%M:%S')

    def _para_python(self, valor):
        return datetime.strptime(valor, self.formato)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# flake8: noqa

from .analizador import Analizador, Dicionario
from .campos import CampoBase, CampoString, CampoNumerico, \
    CampoInteiro, CampoData, CampoDataHora, CampoBooleano

from .decoradores import ProxyDeCampo, ProxyConcatenaAteRE

from .cache import Cache



#coding: utf-8
import re
import weakref

from .cache import Cache
from .colecoes import Dicionario
import collections
import logging

logger = logging.getLogger(__name__)


class AuxiliarDeAnalizador(object):
    """
    Classe auxiliar para definir comportamentos que serão adicionados
    em todos os analizadores através de herança múltipla.
    Padrão mix-in.
    """
    def __init__(self):
        self.tem_busca_inicio = hasattr(self, '_inicio')
        self.tem_busca_fim = hasattr(self, '_fim')
        self.inicio_encontrado = not self.tem_busca_inicio
        self.cache = Cache(self.qtd_linhas_cache + 1)
        self.valor_padrao = None
        self._atribuir_analizador_nos_campos()

    def _atribuir_analizador_nos_campos(self):
        """
        Atribui uma referência fraca do analizador para seus campos.
        Não foi utilizada referência forte para não gerar dependência ciclica,
            impedindo a liberação de memória do analizador.
        """
        ref = weakref.ref(self)
        for item in list(self._campos.values()):
            if hasattr(item, 'atribuir_analizador'):
                item.atribuir_analizador(ref)

    def analizar(self, arquivo, codificacao='latin1'):
        for item in self.analizar_arquivo(arquivo, codificacao):
            yield item

    @property
    def tem_retorno(self):
        return hasattr(self, 'retorno') and self.retorno is not None

    def converter_linha(self, linha, codificacao):
        logger.debug('converter_linha (%s): %s', codificacao, linha)
        if codificacao == 'utf-8':
            return linha
        try:
            return linha.decode(codificacao).encode('utf-8')
        except:
            return linha

    def analizar_arquivo(self, arquivo, codificacao='latin1'):
        try:
            while True:
                linha = next(arquivo)
                linha = self.converter_linha(linha, codificacao)
                res = self.analizar_linha(linha)
                if res:
                    yield res
        except StopIteration:
            res = self.finalizar()
            if res:
                yield res

    def analizar_linha(self, linha):
        self.cache.adicionar(linha)

        if self.tem_busca_inicio and not self.inicio_encontrado:
            self.inicio_encontrado = bool(self._inicio.match(linha))

        if self.inicio_encontrado:
            if not self.tem_retorno:
                self.retorno = Dicionario()
            if self.tem_busca_fim:
                self.inicio_encontrado = not bool(self._fim.match(linha))

            for linha in self.cache.consumir():
                for nome, campo in list(self._campos.items()):
                    if nome in self.retorno and \
                            hasattr(campo, 'lista') and not campo.lista:
                        continue
                    valor = campo.analizar_linha(linha)
                    if valor is not None:
                        self.atribuir_valor_ao_retorno(nome, valor)
                        if self.retornar_ao_obter_valor:
                            return self.finalizar_retorno()

            if not self.inicio_encontrado:
                return self.finalizar_retorno()

    def finalizar(self):
        if not self.tem_retorno:
            return None
        if self.retornar_ao_obter_valor:
            return None
        return self.finalizar_retorno()

    def finalizar_retorno(self):
        for nome, campo in list(self._campos.items()):
            if not nome in self.retorno:
                valor = None
                if hasattr(campo, 'finalizar') and \
                        isinstance(campo.finalizar, collections.Callable):
                    valor = campo.finalizar()
                if valor is None:
                    valor = campo.valor_padrao
                if valor is not None:
                    self.atribuir_valor_ao_retorno(nome, valor)

        self.processar_retorno()
        res = self.retorno
        self.retorno = None
        return res

    def atribuir_valor_ao_retorno(self, nome, valor):
        if isinstance(valor, list) and not nome in self.retorno:
            self.retorno[nome] = valor
        elif isinstance(valor, list) and hasattr(self.retorno[nome], 'extend'):
            self.retorno[nome].extend(valor)
        else:
            self.retorno[nome] = valor

    def processar_retorno(self):
        "Permite modificações finais ao objeto sendo retornado"
        pass


class MetaclasseDeAnalizador(type):
    """
    Metaclasse responsável por descobrir os coletores de informações associados
    à um parser.
    """
    def __new__(self, name, bases, attrs):
        if object in bases:
            bases = tuple([c for c in bases if c != object])

        return type.__new__(self, name, bases + (AuxiliarDeAnalizador,), attrs)

    def __init__(cls, name, bases, attrs):
        super(MetaclasseDeAnalizador, cls).__init__(name, bases, attrs)

        cls._campos = {k: v for k, v in list(attrs.items())
                       if hasattr(v, 'analizar_linha')
                       and not isinstance(v, type)}

        cls.adicionar_atributo_re(cls, attrs, 'inicio')
        cls.adicionar_atributo_re(cls, attrs, 'fim')

        if not hasattr(cls, 'qtd_linhas_cache'):
            cls.qtd_linhas_cache = 0

        if not hasattr(cls, 'retornar_ao_obter_valor'):
            cls.retornar_ao_obter_valor = False

        for nome, atributo in list(cls._campos.items()):
            if hasattr(atributo, 'anexar_na_classe'):
                atributo.anexar_na_classe(cls, nome, cls._campos)

    def adicionar_atributo_re(self, cls, atributos, nome):
        if nome in atributos:
            expressao = atributos[nome]
            setattr(cls, '_' + nome, re.compile(expressao))

    @classmethod
    def __prepare__(self, name, bases):
        return Dicionario()


Analizador = MetaclasseDeAnalizador('Analizador', (object,), {})



#!/usr/bin/env python
#coding: utf-8

from distutils.core import setup

with open('README.rst') as f:
    long_description = f.read()

setup(
    name='raspador',
    author='Fernando Macedo',
    author_email='fgmacedo@gmail.com',
    description='Biblioteca para extração de dados em documentos',
    long_description=long_description,
    license='MIT',
    url="http://github.org/fgmacedo/raspador",
    version='0.1.2',
    packages=['raspador'],
)


#coding: utf-8
import re


class ProxyDeCampo(object):
    def __init__(self, campo):
        self.campo = campo

    def __getattr__(self, atributo):
        return getattr(self.campo, atributo)


class ProxyConcatenaAteRE(ProxyDeCampo):
    """
    Proxy que faz cache de linhas recebidas. Quando recebe uma linha para
    análise, envia a linha ao cache, até encontrar um match da linha recebida
    com a expressão regular de término, e então envia o acumulado das linhas
    recebidas para o decorado.
    """
    def __init__(self, campo, uniao, re_fim):
        super(ProxyConcatenaAteRE, self).__init__(campo)
        self.cache = []
        self.uniao = uniao
        self.re_fim = re.compile(re_fim)

    def analizar_linha(self, linha):
        linha = linha.rstrip()
        self.cache.append(linha)
        if self.re_fim.match(linha):
            acumulado = self.uniao(self.cache)
            self.cache = []
            return self.campo.analizar_linha(acumulado)
